#!/usr/bin/env python

"""
RSVP instruction classes.

Copyright (C) 2007, 2008, 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.data import Attr, Const
from micropython.common import Block

def name(attr):
    if isinstance(attr, Attr):
        return attr.name or "<unnamed>"
    elif isinstance(attr, (Block, Const)):
        return attr
    else:
        return attr.full_name() or "<unnamed>"

class Instruction:

    "A generic instruction."

    def __init__(self, attr=None):
        self.attr = attr
        self.input = None
        self.source = None # for storage instructions

    def copy(self):
        return self.__class__(self.attr)

    def __repr__(self):
        if self.attr is not None:
            return "%s(%r)%s" % (self.__class__.__name__, self.attr, self.show_input())
        else:
            return "%s()%s" % (self.__class__.__name__, self.show_input())

    def show_input(self):
        if self.input is not None:
            if self.source is not None:
                return " <- (%r, %r)" % (self.input, self.source)
            else:
                return " <- %r" % self.input
        elif self.source is not None:
            return " <-- %r" % self.source
        else:
            return ""

    def get_operand(self):
        return None

class FrameRelativeInstruction(Instruction):

    "An instruction operating on the current frame."

    def __repr__(self):
        return "%s(%r)%s" % (self.__class__.__name__, self.get_operand(), self.show_input())

    def get_operand(self):
        return self.attr.position

FR = FrameRelativeInstruction

class AddressRelativeInstruction(Instruction):

    "An instruction accessing an object's attribute."

    def __repr__(self):
        position = self.get_operand()
        if position is not None:
            return "%s(%r)%s # %s" % (self.__class__.__name__, position, self.show_input(), name(self.attr))
        else:
            return "%s(%r)%s" % (self.__class__.__name__, name(self.attr), self.show_input())

    def get_operand(self):
        return self.attr.position

AR = AddressRelativeInstruction

class AddressInstruction(Instruction):

    "An instruction loading an address directly."

    def __repr__(self):
        location, position, result = self.get_operands()
        if location is not None:
            return "%s(%r)%s # %r, %r (%s)" % (
                self.__class__.__name__, result, self.show_input(), location, position, name(self.attr))
        elif result is not None:
            return "%s(%r)%s # %s" % (
                self.__class__.__name__, result, self.show_input(), name(self.attr))
        else:
            return "%s(...)%s # %s" % (
                self.__class__.__name__, self.show_input(), name(self.attr))

    def get_operands(self):
        if isinstance(self.attr, Attr):
            position = self.attr.position
            location = self.attr.parent.location

            # NOTE: Unpositioned attributes are handled here.

            if location is not None and position is not None:
                result = location + position + 1
            else:
                location = self.attr.parent.name
                position = self.attr.name
                result = None
            return location, position, result
        else:
            return None, None, self.attr.location

    def get_operand(self):
        return self.get_operands()[-1]

Address = AddressInstruction

class ImmediateInstruction(Instruction):

    "An instruction employing a constant."

    def __repr__(self):
        return "%s(%r)%s" % (self.__class__.__name__, self.attr, self.show_input())

    def get_operand(self):
        return self.attr

Immediate = ImmediateInstruction

# Access to stored constant data.

class LoadConst(Address):           "Load the constant, class, function, module from the specified location."

# Access within an invocation frame.

class LoadName(FR):                 "Load the current value from the given local attribute/variable."
class StoreName(FR):                "Store the source value into the given local attribute/variable."
class LoadTemp(Immediate):          "Load the current value from the given temporary location."
class StoreTemp(Immediate):         "Store the current value into the given temporary location."

# Access to static data.

class LoadAddress(Address):         "Load the current value from the given fixed attribute address."
class StoreAddress(Address):        "Store the source value into the given fixed attribute address."
class LoadAddressContext(Address):  "Load the current value from the given fixed attribute address, making the current value the context."
class LoadAddressContextCond(Address):
                                  """Load the current value from the given fixed attribute address, only making the current value the
                                     context if the attribute is compatible."""
class MakeObject(Immediate):        "Make a new object. There isn't a complementary DropObject."

# Access to address-relative data.

class LoadAttr(AR):                 "Load into the current value the given attribute of the object referenced by the current value."
class StoreAttr(AR):                "Store the source value into the given attribute of the object referenced by the current value."
class LoadAttrIndex(Immediate):     "Load into the current value the attribute of the current value with the given index."
class StoreAttrIndex(Immediate):    "Store the source value into the attribute of the current value with the given index."
class LoadAttrIndexContext(Immediate):
                                  """Load into the current value the attribute of the current value with the given index, making the
                                     current value the context."""
class LoadAttrIndexContextCond(Immediate):
                                  """Load into the current value the attribute of the current value with the given index, only making the
                                     current value the context if the attribute is compatible."""

# Access to object details.

class LoadCallable(Instruction):    "Load the target of an invocation."
class StoreCallable(Instruction):   "Store the source value into the object referenced by the current value."

# Access to invocation frames in preparation.

class MakeFrame(Immediate):         "Make a new invocation frame."
class DropFrame(Instruction):       "Drop an invocation frame."
class RecoverFrame(Instruction):    "Recover the current frame as an invocation frame."
class StoreFrame(Immediate):        "Store the current value as an argument for the parameter with the given position."
class StoreFrameIndex(Immediate):   "Store the source value as an argument of the current value for the parameter with the given index."
class LoadContext(Instruction):     "Load the context of an invocation."
class CheckFrame(Immediate):        "Check the invocation frame and context for the target."
class CheckSelf(Instruction):       "Check the first argument of an invocation against the target."

# Invocation-related instructions, using a special result "register".

class JumpWithFrame(Instruction):   "Jump, adopting the invocation frame, to the current callable."
class ExtendFrame(Immediate):       "Extend the current frame for temporary storage use."
class AdjustFrame(Immediate):       "Adjust the current frame for corrected invocations."
class Return(Instruction):          "Return from a subprogram."
class LoadResult(Instruction):      "Load into the current value a returned value."
class StoreResult(Instruction):     "Store the current value as a value to be returned."

# Branch-related instructions.

class Jump(Address):                "Jump unconditionally."
class JumpIfFalse(Address):         "Jump if the last evaluation gave a false result."
class JumpIfTrue(Address):          "Jump if the last evaluation gave a true result."

# Exception-related instructions, using a special exception "register".

class LoadException(Instruction):   "Load the raised exception."
class StoreException(Instruction):  "Store the current object in the exception register."
class RaiseException(Instruction):  "Raise an exception, jumping to the active handler."
class PushHandler(Address):         "Push an exception handler onto the handler stack."
class PopHandler(Instruction):      "Pop an exception handler from the handler stack."
class CheckException(Instruction):  "Check the raised exception against another."

# Test instructions, operating on the boolean status register.

class TestIdentity(Instruction):    "Test whether the current value is identical to the source value, setting the boolean status."
class TestIdentityAddress(Address): "Test whether the current value is identical to the given address, setting the boolean status."
class InvertBoolean(Instruction):   "Invert the boolean status."

# Instructions which affect the current value.

current_value_instructions = (
    LoadConst, LoadName, LoadTemp,
    LoadAddress, LoadAddressContext, LoadAddressContextCond,
    LoadAttr, LoadAttrIndex, LoadAttrIndexContext, LoadAttrIndexContextCond,
    LoadCallable, LoadContext, LoadResult,
    LoadException, MakeObject
    )

# vim: tabstop=4 expandtab shiftwidth=4
